from abc import abstractmethod

from aiohttp.web import Application
from aiohttp_asgi import ASGIResource
from aiohttp_asgi.resource import ASGIApplicationType

from .aiohttp import AIOHTTPService, AIOHTTPSSLService


def _create_app(asgi_app: ASGIApplicationType) -> Application:
    app = Application()
    asgi_resource = ASGIResource(asgi_app, root_path="/")
    app.router.register_resource(asgi_resource)
    asgi_resource.lifespan_mount(app, startup=True, shutdown=True)
    return app


class ASGIHTTPService(AIOHTTPService):
    __async_required__ = "start", "create_application", "create_asgi_app"

    async def create_asgi_app(self) -> ASGIApplicationType:
        raise NotImplementedError(
            "You must implement "
            '"create_asgi_app" method',
        )

    async def create_application(self) -> Application:
        return _create_app(await self.create_asgi_app())


class ASGIHTTPSSLService(AIOHTTPSSLService):
    __async_required__ = "start", "create_application", "create_asgi_app"

    @abstractmethod
    async def create_asgi_app(self) -> ASGIApplicationType:
        raise NotImplementedError(
            "You must implement "
            '"create_asgi_app" method',
        )

    async def create_application(self) -> Application:
        return _create_app(await self.create_asgi_app())



import logging
import os
import socket
from typing import Any, Optional, Tuple

from aiomisc.entrypoint import entrypoint
from aiomisc.periodic import PeriodicCallback
from aiomisc.service.base import Service
from aiomisc.utils import TimeoutType


log = logging.getLogger(__name__)


def _get_watchdog_interval() -> Optional[TimeoutType]:
    value = os.getenv("WATCHDOG_USEC")
    if value is None:
        return None
    return int(value) / 1000000.


WATCHDOG_INTERVAL: Optional[TimeoutType] = _get_watchdog_interval()


class SDWatchdogService(Service):
    socket: socket.socket
    watchdog_interval: Optional[TimeoutType]
    _watchdog_timer: PeriodicCallback

    async def _send(self, payload: str) -> None:
        try:
            await self.loop.sock_sendall(
                self.socket,
                payload.encode(),
            )
        except (ConnectionError, OSError) as e:
            log.warning("SystemD notify socket communication problem: %r", e)

    async def _post_start(
        self, services: Tuple[Service, ...], **__: Any
    ) -> None:
        if not hasattr(self, "socket"):
            return

        await self._send(f"STATUS=Started {len(services)} services")
        await self._send("READY=1")

    async def _pre_stop(self, *_: Any, **__: Any) -> None:
        if not hasattr(self, "socket"):
            return
        await self._send("STOPPING=1")

    def __init__(
        self, *, watchdog_interval: Optional[TimeoutType] = WATCHDOG_INTERVAL,
        **kwargs: Any
    ):
        self.watchdog_interval = watchdog_interval
        entrypoint.POST_START.connect(self._post_start)
        entrypoint.PRE_STOP.connect(self._pre_stop)
        super().__init__(**kwargs)

    @staticmethod
    def _get_socket_addr() -> Optional[str]:
        addr = os.getenv("NOTIFY_SOCKET")
        if addr is None:
            return None

        if addr[0] == "@":
            addr = "\0" + addr[1:]

        return addr

    async def start(self) -> None:
        addr = self._get_socket_addr()
        if addr is None:
            log.debug(
                "NOTIFY_SOCKET not exported. Skipping service %r", self,
            )
            return None

        self.socket = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
        self.socket.connect(addr)
        self.socket.setblocking(False)

        await self._send("STATUS=starting")

        if self.watchdog_interval is None:
            return

        if self.watchdog_interval != WATCHDOG_INTERVAL:
            watchdog_usec = int(self.watchdog_interval * 1000000)
            await self._send(f"WATCHDOG_USEC={watchdog_usec}")

        self.start_event.set()
        self._watchdog_timer = PeriodicCallback(
            self._send, "WATCHDOG=1",
        )
        # Send notifications twice as often
        self._watchdog_timer.start(self.watchdog_interval / 2)

        # Removing signals from entrypoint factory because the currently
        # running entrypoint instance has been cloned the signals.
        entrypoint.POST_START.disconnect(self._post_start)
        entrypoint.PRE_STOP.disconnect(self._pre_stop)

    async def stop(self, exception: Exception = None) -> Any:
        await self._watchdog_timer.stop()



import logging
import re
from types import MappingProxyType
from typing import Iterable

from aiocarbon.protocol.pickle import PickleClient  # type: ignore
from aiocarbon.protocol.tcp import TCPClient  # type: ignore
from aiocarbon.protocol.udp import UDPClient  # type: ignore
from aiocarbon.setup import set_client  # type: ignore
from aiocarbon.storage import TotalStorage  # type: ignore

from aiomisc.periodic import PeriodicCallback
from aiomisc.service import Service


log = logging.getLogger(__name__)


def strip_carbon_ns(string: str) -> str:
    return re.sub(r"[^\w\d\-]+", "_", string).strip("_").lower()


PROTOCOLS = MappingProxyType({
    "udp": UDPClient,
    "tcp": TCPClient,
    "pickle": PickleClient,
})


class CarbonSender(Service):
    host: str = "127.0.0.1"
    port: int = 2003
    send_interval: int = 5
    protocol: str = "udp"
    namespace: Iterable[str] = ("",)
    storage = TotalStorage
    _handle: PeriodicCallback

    async def start(self) -> None:
        namespace = ".".join(
            strip_carbon_ns(item) for item in self.namespace
        )

        client = PROTOCOLS[self.protocol](
            self.host,
            self.port,
            namespace=namespace,
            storage=self.storage(),
        )

        set_client(client)

        self._handle = PeriodicCallback(client.send)
        self._handle.start(self.send_interval, loop=self.loop)
        log.info(
            "Periodic carbon metrics sender started. Send to %s://%s:%d with "
            "interval %rs", self.protocol, self.host, self.port,
            self.send_interval,
        )

    async def stop(self, exc: Exception = None) -> None:
        await self._handle.stop()



import asyncio
import logging
from datetime import datetime, timezone
from functools import partial
from typing import Any, Awaitable, Callable, Optional, Tuple, Type, Union

from croniter import croniter

from .compat import EventLoopMixin
from .recurring import RecurringCallback


log = logging.getLogger(__name__)


class CronCallback(EventLoopMixin):
    """
    .. note::

        When the cron function executes longer then execution interval a
        next call will be skipping and warning will be logged.

    """

    __slots__ = ("_recurring_cb", "_task") + EventLoopMixin.__slots__

    def __init__(
        self,
        coroutine_func: Callable[..., Union[Any, Awaitable[Any]]],
        *args: Any, **kwargs: Any
    ):
        self._recurring_cb = RecurringCallback(
            coroutine_func, *args, **kwargs
        )
        self._task: Optional[asyncio.Task] = None

    @staticmethod
    def get_next(cron: croniter, _: RecurringCallback) -> float:
        timestamp = datetime.now(timezone.utc).timestamp()
        next_date = cron.get_next(float, timestamp)
        return next_date - timestamp

    def start(
        self,
        spec: str,
        loop: asyncio.AbstractEventLoop = None,
        *, shield: bool = False,
        suppress_exceptions: Tuple[Type[Exception], ...] = ()
    ) -> None:
        if self._task and not self._task.done():
            raise asyncio.InvalidStateError

        self._loop = loop

        # noinspection PyAttributeOutsideInit
        strategy = partial(self.get_next, croniter(spec))

        self._task = self._recurring_cb.start(
            strategy=strategy,
            loop=loop,
            shield=shield,
            suppress_exceptions=suppress_exceptions,
        )

    def stop(self) -> asyncio.Future:
        if self._task is None:
            task = self.loop.create_future()
            task.set_exception(RuntimeError("Callback not started"))
            return task

        elif not self._task.done():
            self._task.cancel()
        return self._task

    def __repr__(self) -> str:
        return "<%s(%s)>" % (self.__class__.__name__, self._recurring_cb.name)



import logging
import tracemalloc
from enum import Enum
from typing import Any, List, Union

from ..periodic import PeriodicCallback
from ..service import Service
from ..thread_pool import threaded


log = logging.getLogger(__name__)


class GroupBy(Enum):
    lineno = "lineno"
    filename = "filename"
    traceback = "traceback"


class MemoryTracer(Service):
    _tracer: PeriodicCallback
    _log: logging.Logger
    _snapshot_on_start: tracemalloc.Snapshot

    logger = log.info

    interval: int = 5
    top_results: int = 20

    group_by: GroupBy = GroupBy.lineno

    STAT_FORMAT: str = (
        "%(count)8s | "
        "%(count_diff)8s | "
        "%(size)8s | "
        "%(size_diff)8s | "
        "%(traceback)s\n"
    )

    async def start(self) -> None:
        log.warning("Start memory tracer")
        tracemalloc.start()

        self._tracer = PeriodicCallback(self.show_stats)
        self._log = log.getChild(str(id(self)))

        self._snapshot_on_start = self.take_snapshot()
        self._tracer.start(self.interval)

    @staticmethod
    def humanize(num: Union[int, float], suffix: str = "B") -> str:
        for unit in ("", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"):
            if abs(num) < 1024.0:
                return "%3.1f%s%s" % (num, unit, suffix)
            num /= 1024.0
        return "%.1f%s%s" % (num, "Yi", suffix)

    @staticmethod
    def take_snapshot() -> tracemalloc.Snapshot:
        return tracemalloc.take_snapshot()

    def compare_snapshot(
        self, snapshot_from: tracemalloc.Snapshot,
        snapshot_to: tracemalloc.Snapshot,
    ) -> List[tracemalloc.StatisticDiff]:
        return snapshot_to.compare_to(snapshot_from, self.group_by.value)

    def log_diff(self, diff: List[tracemalloc.StatisticDiff]) -> None:
        results = self.STAT_FORMAT % {
            "count": "Objects",
            "count_diff": "Obj.Diff",
            "size": "Memory",
            "size_diff": "Mem.Diff",
            "traceback": "Traceback",
        }
        for stat in diff[:self.top_results]:
            results += self.STAT_FORMAT % {
                "count": stat.count,
                "count_diff": stat.count_diff,
                "size": self.humanize(stat.size),
                "size_diff": self.humanize(stat.size_diff),
                "traceback": stat.traceback,
            }

        self.logger("Top memory usage:\n%s", results)

    @threaded
    def show_stats(self) -> Any:
        if self._snapshot_on_start is None:
            raise RuntimeError

        differences = self.compare_snapshot(
            self._snapshot_on_start,
            self.take_snapshot(),
        )

        return self.log_diff(differences)

    async def stop(self, exception: Exception = None) -> None:
        tracemalloc.stop()



import asyncio
import inspect
import threading
from collections import deque
from concurrent.futures import Executor
from types import TracebackType
from typing import (
    Any, AsyncIterator, Awaitable, Callable, Deque, Generator, NoReturn,
    Optional, Type, TypeVar,
)
from weakref import finalize

from aiomisc.compat import EventLoopMixin
from aiomisc.counters import Statistic


T = TypeVar("T")
R = TypeVar("R")

GenType = Generator[T, R, None]
FuncType = Callable[[], GenType]


class ChannelClosed(RuntimeError):
    pass


class FromThreadChannel(EventLoopMixin):
    __slots__ = (
        "maxsize", "queue", "__close_event", "_max_size",
        "__write_condition", "__read_event", "__get_lock",
    ) + EventLoopMixin.__slots__

    def __init__(self, maxsize: int, loop: asyncio.AbstractEventLoop):
        self.queue: Deque[Any] = deque()
        self._max_size = maxsize
        self._loop = loop
        self.__close_event = threading.Event()
        self.__write_condition = threading.Condition()
        self.__read_event = asyncio.Event()
        self.__get_lock = asyncio.Lock()

    def __notify_readers(self) -> None:
        self.loop.call_soon_threadsafe(self.__read_event.set)

    def __notify_writers(self) -> None:
        with self.__write_condition:
            self.__write_condition.notify_all()

    def close(self) -> None:
        if self.is_closed:
            return

        self.__close_event.set()
        self.__notify_readers()
        self.__notify_writers()

    @property
    def is_overflow(self) -> bool:
        if self._max_size > 0:
            return len(self.queue) >= self._max_size
        return False

    @property
    def is_empty(self) -> bool:
        return len(self.queue) == 0

    @property
    def is_closed(self) -> bool:
        return self.__close_event.is_set()

    def __enter__(self) -> "FromThreadChannel":
        return self

    def __exit__(
        self, exc_type: Type[Exception],
        exc_val: Exception, exc_tb: TracebackType,
    ) -> None:
        self.close()

    def put(self, item: Any) -> None:
        def predicate() -> bool:
            return self.is_closed or not self.is_overflow

        with self.__write_condition:
            self.__write_condition.wait_for(predicate)

            if self.is_closed:
                raise ChannelClosed

            self.queue.append(item)
            self.__notify_readers()

    async def get(self) -> Any:
        async with self.__get_lock:
            if self.is_closed:
                if self.is_empty:
                    raise ChannelClosed
                return self.queue.popleft()

            while self.is_empty:
                await self.__read_event.wait()
                if self.is_closed and self.is_empty:
                    raise ChannelClosed
            try:
                return self.queue.popleft()
            finally:
                if self.is_empty and not self.is_closed:
                    self.__read_event.clear()
                self.__notify_writers()


class IteratorWrapperStatistic(Statistic):
    started: int
    queue_size: int
    queue_length: int
    yielded: int
    enqueued: int


class IteratorWrapper(AsyncIterator, EventLoopMixin):
    __slots__ = (
        "__channel",
        "__close_event",
        "__gen_func",
        "__gen_task",
        "_statistic",
        "executor",
    ) + EventLoopMixin.__slots__

    def __init__(
        self, gen_func: FuncType,
        loop: asyncio.AbstractEventLoop = None,
        max_size: int = 0, executor: Executor = None,
        statistic_name: Optional[str] = None,
    ):

        self._loop = loop
        self.executor = executor

        self.__close_event = asyncio.Event()
        self.__channel: FromThreadChannel = FromThreadChannel(
            maxsize=max_size, loop=self.loop,
        )
        self.__gen_task: Optional[asyncio.Future] = None
        self.__gen_func: Callable = gen_func
        self._statistic = IteratorWrapperStatistic(statistic_name)
        self._statistic.queue_size = max_size

    @property
    def closed(self) -> bool:
        return self.__channel.is_closed

    @staticmethod
    def __throw(_: Any) -> NoReturn:
        pass

    def _in_thread(self) -> None:
        self._statistic.started += 1
        with self.__channel:
            try:
                gen = iter(self.__gen_func())

                throw = self.__throw
                if inspect.isgenerator(gen):
                    throw = gen.throw   # type: ignore

                while not self.closed:
                    item = next(gen)
                    try:
                        self.__channel.put((item, False))
                    except Exception as e:
                        throw(e)
                        self.__channel.close()
                        break
                    finally:
                        del item

                    self._statistic.enqueued += 1
            except StopIteration:
                return
            except Exception as e:
                if self.closed:
                    return
                self.__channel.put((e, True))
            finally:
                self._statistic.started -= 1
                self.loop.call_soon_threadsafe(self.__close_event.set)

    def close(self) -> Awaitable[None]:
        self.__channel.close()
        return asyncio.ensure_future(self.wait_closed())

    async def wait_closed(self) -> None:
        await self.__close_event.wait()
        if self.__gen_task:
            await asyncio.gather(self.__gen_task, return_exceptions=True)

    def __aiter__(self) -> AsyncIterator[Any]:
        if self.__gen_task is None:
            gen_task = self.loop.run_in_executor(
                self.executor, self._in_thread,
            )
            if gen_task is None:
                raise RuntimeError("Iterator task was not created")
            self.__gen_task = gen_task
        return IteratorProxy(self, self.close)

    async def __anext__(self) -> Awaitable[T]:
        try:
            item, is_exc = await self.__channel.get()
        except ChannelClosed:
            await self.wait_closed()
            raise StopAsyncIteration

        if is_exc:
            await self.close()
            raise item from item

        self._statistic.yielded += 1
        return item

    async def __aenter__(self) -> "IteratorWrapper":
        return self

    async def __aexit__(
        self, exc_type: Any, exc_val: Any,
        exc_tb: Any,
    ) -> None:
        if self.closed:
            return

        await self.close()


class IteratorProxy(AsyncIterator):
    def __init__(
        self, iterator: AsyncIterator,
        finalizer: Callable[[], Any],
    ):
        self.__iterator = iterator
        finalize(self, finalizer)

    def __anext__(self) -> Awaitable[Any]:
        return self.__iterator.__anext__()



import asyncio
import logging
import os
import sys
from concurrent.futures import Executor
from typing import (
    Any, Callable, Coroutine, MutableSet, Optional, TypeVar, Union,
)
from weakref import WeakSet

import aiomisc_log
from aiomisc_log import LogLevel

from .compat import event_loop_policy, set_current_loop
from .context import Context, get_context
from .log import LogFormat, basic_config
from .service import Service
from .signal import Signal
from .utils import cancel_tasks, create_default_event_loop


ExecutorType = Executor
T = TypeVar("T")
log = logging.getLogger(__name__)


if sys.version_info < (3, 7):
    asyncio_all_tasks = asyncio.Task.all_tasks
    asyncio_current_task = asyncio.Task.current_task
else:
    asyncio_all_tasks = asyncio.all_tasks
    asyncio_current_task = asyncio.current_task


def _get_env_bool(name: str, default: str) -> bool:
    enable_variants = {"1", "enable", "enabled", "on", "true", "yes"}
    return os.getenv(name, default).lower() in enable_variants


def _get_env_convert(name: str, converter: Callable[..., T], default: T) -> T:
    value = os.getenv(name)
    if value is None:
        return default
    return converter(value)


class Entrypoint:
    DEFAULT_LOG_LEVEL: str = os.getenv(
        "AIOMISC_LOG_LEVEL", LogLevel.default(),
    )
    DEFAULT_LOG_FORMAT: str = os.getenv(
        "AIOMISC_LOG_FORMAT", LogFormat.default(),
    )

    DEFAULT_AIOMISC_DEBUG: bool = _get_env_bool("AIOMISC_DEBUG", "0")
    DEFAULT_AIOMISC_LOG_CONFIG: bool = _get_env_bool(
        "AIOMISC_LOG_CONFIG", "1",
    )
    DEFAULT_AIOMISC_LOG_FLUSH: float = _get_env_convert(
        "AIOMISC_LOG_FLUSH", float, 0.2,
    )
    DEFAULT_AIOMISC_BUFFERING: bool = _get_env_bool(
        "AIOMISC_LOG_BUFFERING", "1",
    )
    DEFAULT_AIOMISC_BUFFER_SIZE: int = _get_env_convert(
        "AIOMISC_LOG_BUFFER", int, 1024,
    )
    DEFAULT_AIOMISC_POOL_SIZE: Optional[int] = _get_env_convert(
        "AIOMISC_POOL_SIZE", int, None,
    )

    PRE_START = Signal()
    POST_STOP = Signal()
    POST_START = Signal()
    PRE_STOP = Signal()

    async def _start(self) -> None:
        if self.log_config:
            basic_config(
                level=self.log_level,
                log_format=self.log_format,
                buffered=self.log_buffering,
                loop=self.loop,
                buffer_size=self.log_buffer_size,
                flush_interval=self.log_flush_interval,
            )

        set_current_loop(self.loop)

        for signal in (
            self.pre_start, self.post_stop,
            self.pre_stop, self.post_start,
        ):
            signal.freeze()

        await self.pre_start.call(entrypoint=self, services=self.services)

        await asyncio.gather(
            *[self._start_service(svc) for svc in self.services],
        )

        await self.post_start.call(entrypoint=self, services=self.services)

    def __init__(
        self, *services: Service, loop: asyncio.AbstractEventLoop = None,
        pool_size: int = None,
        log_level: Union[int, str] = DEFAULT_LOG_LEVEL,
        log_format: Union[str, LogFormat] = DEFAULT_LOG_FORMAT,
        log_buffering: bool = DEFAULT_AIOMISC_BUFFERING,
        log_buffer_size: int = DEFAULT_AIOMISC_BUFFER_SIZE,
        log_flush_interval: float = DEFAULT_AIOMISC_LOG_FLUSH,
        log_config: bool = DEFAULT_AIOMISC_LOG_CONFIG,
        policy: asyncio.AbstractEventLoopPolicy = event_loop_policy,
        debug: bool = DEFAULT_AIOMISC_DEBUG
    ):

        """

        :param debug: set debug to event loop
        :param loop: loop
        :param services: Service instances which will be starting.
        :param pool_size: thread pool size
        :param log_level: Logging level which will be configured
        :param log_format: Logging format which will be configured
        :param log_buffer_size: Buffer size for logging
        :param log_flush_interval: interval in seconds for flushing logs
        :param log_config: if False do not configure logging
        """

        self._debug = debug
        self._loop = loop
        self._loop_owner = False
        self._tasks: MutableSet[asyncio.Task] = WeakSet()
        self._thread_pool: Optional[ExecutorType] = None
        self._closing: Optional[asyncio.Event] = None

        self.ctx: Optional[Context] = None
        self.log_buffer_size = log_buffer_size
        self.log_buffering = log_buffering
        self.log_config = log_config
        self.log_flush_interval = log_flush_interval
        self.log_format = log_format
        self.log_level = log_level
        self.policy = policy
        self.pool_size = pool_size
        self.services = services
        self.shutting_down = False
        self.pre_start = self.PRE_START.copy()
        self.post_start = self.POST_START.copy()
        self.pre_stop = self.PRE_STOP.copy()
        self.post_stop = self.POST_STOP.copy()

        if self.log_config:
            aiomisc_log.basic_config(
                level=self.log_level,
                log_format=self.log_format,
            )

        if self._loop is not None:
            set_current_loop(self._loop)

    async def closing(self) -> None:
        # Lazy initialization because event loop might be not exists
        if self._closing is None:
            self._closing = asyncio.Event()

        await self._closing.wait()

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        if self._loop is None:
            self._loop, self._thread_pool = create_default_event_loop(
                pool_size=self.pool_size,
                policy=self.policy,
                debug=self._debug,
            )
            self._loop_owner = True
            set_current_loop(self._loop)
        return self._loop

    def __del__(self) -> None:
        if self._loop and self._loop.is_closed():
            return

        if self._loop_owner and self._loop is not None:
            self._loop.close()

    def __enter__(self) -> asyncio.AbstractEventLoop:
        self.loop.run_until_complete(self.__aenter__())
        return self.loop

    def __exit__(
        self, exc_type: Any, exc_val: Any, exc_tb: Any,
    ) -> None:
        if self.loop.is_closed():
            return

        if self.log_config:
            basic_config(
                level=self.log_level,
                log_format=self.log_format,
                buffered=False,
            )

        self.loop.run_until_complete(
            self.__aexit__(exc_type, exc_val, exc_tb),
        )

        if self._loop_owner and self._loop is not None:
            self._loop.close()

    async def __aenter__(self) -> "Entrypoint":
        if self._loop is None:
            # When __aenter__ called without __enter__
            self._loop = asyncio.get_running_loop()

        self.ctx = Context(loop=self.loop)
        await self._start()
        return self

    async def __aexit__(
        self, exc_type: Any, exc_val: Any, exc_tb: Any,
    ) -> None:
        await self.pre_stop.call(entrypoint=self)

        try:
            if self.loop.is_closed():
                return

            await self.graceful_shutdown(exc_val)
            self.shutting_down = True
        finally:
            if self.ctx:
                self.ctx.close()

            if self._thread_pool:
                self._thread_pool.shutdown()

    async def _start_service(
        self, svc: Service,
    ) -> None:
        svc.set_loop(self.loop)

        start_task, ev_task = map(
            asyncio.ensure_future, (
                svc.start(), svc.start_event.wait(),
            ),
        )

        await asyncio.wait(
            (start_task, ev_task),
            return_when=asyncio.FIRST_COMPLETED,
        )

        self.loop.call_soon(svc.start_event.set)
        await ev_task

        if start_task.done():
            # raise an Exception when failed
            await start_task
            return
        else:
            self._tasks.add(start_task)

        return None

    async def _cancel_background_tasks(self) -> None:
        tasks = asyncio_all_tasks(self._loop)
        current_task = asyncio_current_task(self.loop)
        await cancel_tasks(task for task in tasks if task is not current_task)

    async def graceful_shutdown(self, exception: Exception) -> None:
        if self._closing:
            self._closing.set()

        tasks = []
        for svc in self.services:
            try:
                coro = svc.stop(exception)
            except TypeError as e:
                log.warning(
                    "Failed to stop service %r:\n%r", svc, e,
                )
                log.debug("Service stop failed traceback", exc_info=True)
            else:
                tasks.append(asyncio.shield(coro))

        if tasks:
            await asyncio.gather(*tasks, return_exceptions=True)

        await cancel_tasks(set(self._tasks))

        await self.post_stop.call(entrypoint=self)

        if self._loop_owner:
            await self._cancel_background_tasks()
        await self.loop.shutdown_asyncgens()


entrypoint = Entrypoint


def run(
    coro: Coroutine[None, Any, T],
    *services: Service,
    **kwargs: Any
) -> T:
    with entrypoint(*services, **kwargs) as loop:
        return loop.run_until_complete(coro)


__all__ = ("entrypoint", "Entrypoint", "get_context", "run")



from .color import color_formatter
from .journald import journald_formatter
from .json import json_handler
from .rich import rich_formatter


__all__ = (
    "color_formatter", "json_handler", "rich_formatter", "journald_formatter",
)



import logging
import logging.handlers
import os
import sys
from types import TracebackType
from typing import Any, Callable, Optional, Type, Union

from .enum import LogFormat, LogLevel
from .formatter import (
    color_formatter, journald_formatter, json_handler, rich_formatter,
)


LOG_LEVEL: Optional[Any] = None
LOG_FORMAT: Optional[Any] = None

try:
    import contextvars
    LOG_LEVEL = contextvars.ContextVar("LOG_LEVEL", default=logging.INFO)
    LOG_FORMAT = contextvars.ContextVar(
        "LOG_FORMAT", default=LogFormat.default(),
    )
except ImportError:
    pass


DEFAULT_FORMAT = "%(levelname)s:%(name)s:%(message)s"


def create_logging_handler(
    log_format: LogFormat = LogFormat.color,
    date_format: str = None, **kwargs: Any
) -> logging.Handler:

    if LOG_FORMAT is not None:
        LOG_FORMAT.set(log_format)

    handler: logging.Handler

    if log_format == LogFormat.stream:
        handler = logging.StreamHandler()
        if date_format and date_format is not Ellipsis:
            formatter = logging.Formatter(
                "%(asctime)s " + DEFAULT_FORMAT, datefmt=date_format,
            )
        else:
            formatter = logging.Formatter(DEFAULT_FORMAT)

        handler.setFormatter(formatter)
        return handler
    elif log_format == LogFormat.json:
        return json_handler(date_format=date_format, **kwargs)
    elif log_format == LogFormat.color:
        return color_formatter(date_format=date_format, **kwargs)
    elif log_format == LogFormat.journald:
        return journald_formatter()
    elif log_format == LogFormat.rich:
        return rich_formatter(date_format=date_format, **kwargs)
    elif log_format == LogFormat.rich_tb:
        return rich_formatter(
            date_format=date_format,
            rich_tracebacks=True,
            **kwargs
        )
    elif log_format == LogFormat.syslog:
        if date_format:
            sys.stderr.write("Can not apply \"date_format\" for syslog\n")
            sys.stderr.flush()

        formatter = logging.Formatter("%(message)s")

        if os.path.exists("/dev/log"):
            handler = logging.handlers.SysLogHandler(address="/dev/log")
        else:
            handler = logging.handlers.SysLogHandler()

        handler.setFormatter(formatter)
        return handler
    elif log_format == LogFormat.plain:
        handler = logging.StreamHandler()
        formatter = logging.Formatter("%(message)s")
        handler.setFormatter(formatter)
        return handler

    raise NotImplementedError


HandlerWrapperType = Callable[[logging.Handler], logging.Handler]


def pass_wrapper(handler: logging.Handler) -> logging.Handler:
    return handler


class UnhandledHookBase:
    __slots__ = "logger",

    LOGGER_NAME: str = "unhandled"
    logger: logging.Logger

    def __init__(self) -> None:
        self.logger = logging.getLogger().getChild(self.LOGGER_NAME)
        self.logger.propagate = False

    def set_handler(self, handler: logging.Handler) -> None:
        self.logger.handlers.clear()
        self.logger.handlers.append(handler)


class UnhandledHook(UnhandledHookBase):
    MESSAGE: str = "Unhandled exception"

    def __call__(
        self,
        exc_type: Type[BaseException],
        exc_value: BaseException,
        exc_traceback: TracebackType,
    ) -> None:
        self.logger.exception(
            self.MESSAGE, exc_info=(exc_type, exc_value, exc_traceback),
        )


def basic_config(
    level: Union[int, str] = logging.INFO,
    log_format: Union[str, LogFormat] = LogFormat.color,
    handler_wrapper: HandlerWrapperType = pass_wrapper,
    **kwargs: Any
) -> None:

    if isinstance(level, str):
        level = LogLevel[level]

    logging.basicConfig()
    logger = logging.getLogger()
    logger.handlers.clear()

    if isinstance(log_format, str):
        log_format = LogFormat[log_format]

    raw_handler = create_logging_handler(log_format, **kwargs)
    unhandled_hook = UnhandledHook()
    sys.excepthook = unhandled_hook     # type: ignore

    handler = handler_wrapper(raw_handler)

    if LOG_LEVEL is not None:
        LOG_LEVEL.set(level)

    # noinspection PyArgumentList
    logging.basicConfig(
        level=int(level),
        handlers=[handler],
    )

    unhandled_hook.set_handler(raw_handler)


__all__ = (
    "LogFormat",
    "LogLevel",
    "basic_config",
    "LOG_FORMAT",
    "LOG_LEVEL",
)
